unit QuickBitmap;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, Menus, math, clipbrd, StdCtrls, pngimage, Buttons, shellapi,
  jpeg;

type
  TRGB = record
    b,g,r:byte;
  end;
  PRGB = ^TRGB;
  TRGBArray = array[0..0] of TRGB;
  PRGBArray = ^TRGBArray;

type
  TFloatRGB = record
    b,g,r:single;
  end;

type
  TLagrangePolynom = array [0..3] of single;

type
  TQuickBitmap = class
    private
      bitmap:TBitmap;
      origin:PByteArray;
      LineLength:integer;
      LX:array of TLagrangePolynom;
      LY:array of TLagrangePolynom;
    public
      width:integer;
      height:integer;
      constructor create(aWidth:integer; aHeight:integer) overload;
      constructor create(aBitmap:TBitmap) overload;
      destructor destroy;
      function GetBitmap:TBitmap;
      function GetPixel(x:integer; y:integer):TRGB;
      procedure SetPixel(x:integer; y:integer; color:TRGB);
      function QuickBicubicResize(xscale:integer; yscale:integer):TQuickBitmap;
  end;

implementation

{ TQuickBitmap }

//vytvori bitmapu
constructor TQuickBitmap.create(aBitmap: TBitmap);
begin
  bitmap:=TBitmap.Create();
  bitmap.PixelFormat:=pf24bit;
  if((aBitmap<>nil) and (aBitmap.Width>1) and (aBitmap.Height>1)) then
  begin
    bitmap.Width:=aBitmap.Width;
    bitmap.Height:=aBitmap.Height;
    bitmap.Canvas.Draw(0,0,aBitmap);
  end
  else
  begin
    bitmap.Width:=100;
    bitmap.Height:=100;
  end;
  width:=bitmap.Width;
  height:=bitmap.Height;
  origin:=bitmap.ScanLine[0];
  LineLength:=(Longint(bitmap.ScanLine[1]) - Longint(origin));
end;


destructor TQuickBitmap.destroy;
begin
  bitmap.Destroy;
  setlength(LX,0);
  setlength(LY,0);
end;

constructor TQuickBitmap.create(aWidth, aHeight: integer);
begin
  bitmap:=TBitmap.Create();
  bitmap.PixelFormat:=pf24bit;
  bitmap.Width:=aWidth;
  bitmap.Height:=aHeight;
  width:=bitmap.Width;
  height:=bitmap.Height;
  origin:=bitmap.ScanLine[0];
  LineLength:=(Longint(bitmap.ScanLine[1]) - Longint(origin));
end;

//zjisti bitmapu
function TQuickBitmap.GetBitmap: TBitmap;
begin
  result:=bitmap;
end;


//ziska pixel
function TQuickBitmap.GetPixel(x:integer; y:integer): TRGB;
begin
  if(x>width-1) then x:=width-1;
  if(x<0) then x:=0;
  if(y>height-1) then y:=height-1;
  if(y<0) then y:=0;
  result.b:=origin^[y*LineLength+x*sizeof(TRGB)];
  result.g:=origin^[y*LineLength+x*sizeof(TRGB)+1];
  result.r:=origin^[y*LineLength+x*sizeof(TRGB)+2];
end;

//rozsiri bitmapu
function TQuickBitmap.QuickBicubicResize(xscale, yscale: integer): TQuickBitmap;
var
  a:integer;
  x:single;
  y:single;
  xi,yi,u,lxi,lyi:integer;
  pix:array[0..3] of TRGB;
  pixelbasex:integer;
  pixelbasey:integer;
  floatrgb:TFloatRGB;
  color:TRGB;
begin
  setlength(LX,xscale);
  setlength(LY,yscale);

  //urci lagrangeovy polynomy
  for a := 0 to xscale-1 do
  begin
    x := (1 / xscale) * a;
    LX[a][0] := ((x - 0) * (x - 1) * (x - 2)) / (-6);
    LX[a][1] := ((x + 1) * (x - 1) * (x - 2)) / (2);
    LX[a][2] := ((x + 1) * (x - 0) * (x - 2)) / (-2);
    LX[a][3] := ((x + 1) * (x - 0) * (x - 1)) / (6);
  end;

  for a := 0 to yscale-1 do
  begin
    y := (1.0 / yscale) * a;
    LY[a][0] := ((y - 0) * (y - 1) * (y - 2)) / (-6);
    LY[a][1] := ((y + 1) * (y - 1) * (y - 2)) / (2);
    LY[a][2] := ((y + 1) * (y - 0) * (y - 2)) / (-2);
    LY[a][3] := ((y + 1) * (y - 0) * (y - 1)) / (6);
  end;

  //vytvori novou bitmapu
  result:=TQuickBitmap.create(width*xscale, height*yscale);

  for xi:=0 to width-1 do
  begin
    for yi:=0 to height-1 do
    begin
      //load pixels
      for u:=0 to 3 do
      begin
        pix[u] := GetPixel(xi + u - 1, yi);
      end;

      PixelBaseX := xi * xscale;
      PixelBaseY := yi * yscale;
      //bicubic enlarge only using x axe
      for lxi:=0 to xscale-1 do
      begin
        floatrgb.r := LX[lxi][0] * pix[0].r + LX[lxi][1] * pix[1].r + LX[lxi][2] * pix[2].r + LX[lxi][3] * pix[3].r;
        floatrgb.g := LX[lxi][0] * pix[0].g + LX[lxi][1] * pix[1].g + LX[lxi][2] * pix[2].g + LX[lxi][3] * pix[3].g;
        floatrgb.b := LX[lxi][0] * pix[0].b + LX[lxi][1] * pix[1].b + LX[lxi][2] * pix[2].b + LX[lxi][3] * pix[3].b;
        color.r := round(min(max(0,floatrgb.r),255));
        color.g := round(min(max(0,floatrgb.g),255));
        color.b := round(min(max(0,floatrgb.b),255));
        result.SetPixel(PixelBaseX + lxi, PixelBaseY, color);
      end;

    end;
  end;

  for xi:=0 to result.width-1 do
  begin
    for yi:=0 to height-1 do
    begin
      //load pixels
      for u:=0 to 3 do
      begin
        pix[u] := result.GetPixel(xi, (yi+u-1) * yscale);
      end;

      PixelBaseX := xi;
      PixelBaseY := yi * yscale;
      //bicubic enlarge only using x axe
      for lyi:=0 to yscale-1 do
      begin
        floatrgb.r := LY[lyi][0] * pix[0].r + LY[lyi][1] * pix[1].r + LY[lyi][2] * pix[2].r + LY[lyi][3] * pix[3].r;
        floatrgb.g := LY[lyi][0] * pix[0].g + LY[lyi][1] * pix[1].g + LY[lyi][2] * pix[2].g + LY[lyi][3] * pix[3].g;
        floatrgb.b := LY[lyi][0] * pix[0].b + LY[lyi][1] * pix[1].b + LY[lyi][2] * pix[2].b + LY[lyi][3] * pix[3].b;
        color.r := round(min(max(0,floatrgb.r),255));
        color.g := round(min(max(0,floatrgb.g),255));
        color.b := round(min(max(0,floatrgb.b),255));
        result.SetPixel(PixelBaseX, PixelBaseY+lyi, color);
      end;

    end;
  end;
end;

//nastavi pixel
procedure TQuickBitmap.SetPixel(x, y: integer; color: TRGB);
begin
  if(x>width-1) then exit;
  if(x<0) then exit;
  if(y>height-1) then exit;
  if(y<0) then exit;
  origin^[y*LineLength+x*sizeof(TRGB)+0]:=color.b;
  origin^[y*LineLength+x*sizeof(TRGB)+1]:=color.g;
  origin^[y*LineLength+x*sizeof(TRGB)+2]:=color.r;
end;

end.
