unit UnitClipForm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, math, Buttons, StdCtrls, pngimage, strutils, shellapi,
  clipbrd;

type
  TClipForm = class(TForm)
    Timer1: TTimer;
    Recordingbtn: TSpeedButton;
    Cancelbtn: TSpeedButton;
    Image1: TImage;
    lbl1: TLabel;
    lbl2: TLabel;
    lbl3: TLabel;
    lbl4: TLabel;
    shape1: TShape;
    Shape2: TShape;
    Shape3: TShape;
    Shape4: TShape;
    CursorImage: TImage;
    CursorMouseDownImage: TImage;
    procedure Timer1Timer(Sender: TObject);
    procedure CancelbtnClick(Sender: TObject);
    procedure RecordingbtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    ClipRect:TRect;
    ClipRgn:THandle;
    //if recording form is shown
    isShowingRecording:boolean;
    //if video is beeing recorded
    isRecording:boolean;
    //if it is video or slow motion
    isSlowMotion:boolean;
    hotkey:integer;
    procedure gosleep(miliseconds:integer);
    procedure positionshapes;
    procedure ShowClipForm;
    procedure HideClipForm;
    procedure ShowRecording(aIsSlowMotion:boolean);
    procedure RecordSlowMotion;
    procedure RecordVideo;
    procedure NormalizeCliprect;
    procedure StopRecording;
    procedure WMHotKey(var Msg: TWMHotKey); message WM_HOTKEY;
  end;

var
  ClipForm: TClipForm;
  path:string;
  uidcounter:longint;

  //last captured video
  LastVideoFilename:string;
  RecordingVideoStatus:integer;
  VideoCodec:string;
  VideoQuality:integer;

  function GetUID:string;
  function RenderVideo(value: pointer):integer;
  procedure DeleteFiles(const DirName: string);

const
  PublishedData:string = '\Data\Published\img';
  PublishedOCR:string = '\Data\Published\ocr';
  PublishedSlowMotion:string = '\Data\Published\slowmotion';
  PublishedVideo:string = '\Data\Published\video';
  TempVideo:string = '\Data\Temp\Video';
  TesseractDir:string = '\Data\TesseractOCR';

  NotRecordingVideo:integer = 0;
  RecordingVideo:integer = 10;
  RenderingVideo:integer = 20;
  RenderingVideoAccepted:integer = 30;
  VideoRendered:integer = 40;

implementation

function GetUID: string;
begin
  result:='-'+rightstr(inttostr(UIDCounter+1000000),4);
  UIDCounter:=UIDCounter+1;
  result:=formatdatetime('yyyy-mm-dd-hh-nn-ss',now)+result;
end;

procedure DeleteFiles(const DirName: string);
var
  FileOp: TSHFileOpStruct;
begin
  //delete files in directory
  FillChar(FileOp, SizeOf(FileOp), 0);
  FileOp.wFunc := FO_DELETE;
  FileOp.pFrom := PChar(DirName+#0);//double zero-terminated
  FileOp.fFlags := FOF_SILENT or FOF_NOERRORUI or FOF_NOCONFIRMATION or
                   FOF_FILESONLY;
  SHFileOperation(FileOp);
end;

function RenderVideo(value: pointer):integer;
var StartInfo : TStartupInfo;
    ProcInfo : TProcessInformation;
    CreateOK : Boolean;
    Codec:string;
    cmdline:string;
begin
try
  //render video
  result:=0;
  RecordingVideoStatus:=RenderingVideo;
  LastVideoFilename:=GetUID;
  if(VideoCodec='') then VideoCodec:='wmv';
  LastVideoFilename:=path+publishedvideo+'\'+LastVideoFilename+'.'+VideoCodec;

  //set codec
  Codec:='-vcodec wmv2';
  if VideoCodec='wmv' then Codec:='-vcodec wmv2';
  if VideoCodec='mp4' then Codec:='-vcodec mpeg4';
  if VideoCodec='flv' then Codec:='-vcodec flv';
  if VideoCodec='mov' then Codec:='';
  if VideoCodec='webm' then Codec:='-f webm -vcodec libvpx';
  //set video quality if it was not set before
  VideoQuality:=min(max(5,VideoQuality),40);

  //start ffmpeg
  FillChar(StartInfo,SizeOf(TStartupInfo), 0);
  FillChar(ProcInfo,SizeOf(TProcessInformation), 0);
  StartInfo.cb := SizeOf(TStartupInfo);
  StartInfo.wShowWindow:=sw_hide;
  cmdline:=path+TesseractDir+'\ff.exe -r 12 -f image2 -i "'+
           path+TempVideo+'\%06d.png" -y '+Codec+' -qscale '+
           inttostr(VideoQuality)+' -r 12 "'+LastVideoFilename+'"';
  clipboard.AsText:=cmdline;
  CreateOK := CreateProcess(nil, PChar(cmdline),
                            nil, nil,False,
                            CREATE_NEW_PROCESS_GROUP or NORMAL_PRIORITY_CLASS,
                            nil, nil, StartInfo, ProcInfo);
  //wait for the result
  if CreateOK then
    while WaitForSingleObject(ProcInfo.hProcess, 50)<>0 do
    begin
      application.ProcessMessages;
      sleep(20);
    end;
  CloseHandle(ProcInfo.hProcess);
  CloseHandle(ProcInfo.hThread);
finally
  RecordingVideoStatus:=VideoRendered;
end;
end;

{$R *.dfm}

{ TClipForm }

procedure TClipForm.CancelbtnClick(Sender: TObject);
begin
  isShowingRecording:=false;
  HideClipForm;
end;

procedure TClipForm.gosleep(miliseconds: integer);
var a:integer;
begin
  for a := 1 to floor(miliseconds/20) do
  begin
    sleep(20);
    application.ProcessMessages;
  end;
end;

procedure TClipForm.HideClipForm;
begin
  visible:=false;
  showwindow(handle,sw_hide);
end;

procedure TClipForm.NormalizeCliprect;
var newrect:TRect;
begin
  //normalize clip rect
  newrect.Left:=min(ClipRect.Left,ClipRect.Right);
  newrect.Top:=min(ClipRect.Top,ClipRect.Bottom);
  newrect.Right:= max(ClipRect.Left,ClipRect.Right);
  newrect.Bottom:=max(ClipRect.Top,ClipRect.Bottom);
  cliprect:=newrect;
end;

procedure TClipForm.positionshapes;
begin
  normalizecliprect;
  //set position of the form
  self.Left:=ClipRect.Left-2;
  self.Top:=ClipRect.Top-2;
  self.Width:= ClipRect.Right-self.Left+4;
  self.Height:=ClipRect.Bottom-self.Top+4;
  //set border
  shape1.SetBounds(0,-2,2,height+4);
  shape2.SetBounds(0,0,width,2);
  shape3.SetBounds(width-2,-2,2,height+4);
  shape4.SetBounds(0,height-2,width,2);
end;

procedure TClipForm.ShowClipForm;
var
  rgn1:THandle;
  rgn2:THandle;
begin
  positionshapes;

  rgn1:=createrectrgn(0,0,width,height);
  rgn2:=createrectrgn(2,2,width-2,height-2);
  ClipRgn:=rgn1;
  combinergn(ClipRgn,rgn1,rgn2,RGN_DIFF);
  setwindowrgn(handle,ClipRgn,true);

  visible:=true;
  showwindow(handle,sw_show);
end;

procedure TClipForm.ShowRecording(aIsSlowMotion: boolean);
var
  rgn1:THandle;
  rgn2:THandle;
  rgn3:THandle;
begin
  //position border shape and show window
  positionshapes;

  //display error message
  if((width<160) or (height<120)) then
  begin
    showmessage('If you want to record video then you have to define larger rectangle.'+
                'Resolution width x height must be higher than 160 x 120.');
    exit;
  end;

  //create rgn
  rgn1:=createrectrgn(0,0,width,height);
  rgn2:=createrectrgn(2,2,width-2,height-2);
  rgn3:=createrectrgn(2,2,152,102);
  ClipRgn:=rgn1;
  combinergn(ClipRgn,rgn1,rgn2,RGN_DIFF);
  combinergn(ClipRgn,ClipRgn,rgn3,RGN_OR);
  setwindowrgn(handle,ClipRgn,true);

  visible:=true;
  showwindow(handle,sw_show);
  isShowingRecording:=true;
  isSlowMotion:=aIsSlowMotion;
end;

procedure TClipForm.StopRecording;
begin
  isRecording:=false;
end;

procedure TClipForm.RecordingbtnClick(Sender: TObject);
begin
  //hide components
  lbl1.Visible:=false;
  lbl2.Visible:=false;
  lbl3.Visible:=false;
  Recordingbtn.Visible:=false;
  Cancelbtn.Visible:=false;

  lbl4.Caption:='Wait';
  lbl4.SetBounds(image1.Left,image1.Top,image1.Width,image1.Height);
  lbl4.Visible:=true;
  application.ProcessMessages;
  gosleep(500);

  //delete previous files
  DeleteFiles(path+PublishedSlowMotion+'\*.*');
  DeleteFiles(path+TempVideo+'\*.*');

  //start countdown
  lbl4.Caption:='3';
  lbl4.SetBounds(image1.Left,image1.Top,image1.Width,image1.Height);
  lbl4.Visible:=true;
  application.ProcessMessages;
  gosleep(1000);
  lbl4.Caption:='2';
  lbl4.SetBounds(image1.Left,image1.Top,image1.Width,image1.Height);
  application.ProcessMessages;
  gosleep(1000);
  lbl4.Caption:='1';
  lbl4.SetBounds(image1.Left,image1.Top,image1.Width,image1.Height);
  application.ProcessMessages;
  gosleep(1000);
  lbl4.Visible:=false;

  //show standard clip form
  ShowClipForm;

  //show components
  lbl1.Visible:=true;
  lbl2.Visible:=true;
  lbl3.Visible:=true;
  Recordingbtn.Visible:=true;
  Cancelbtn.Visible:=true;

  if(isSlowMotion) then
    RecordSlowMotion
  else
    RecordVideo;
end;

procedure TClipForm.RecordSlowMotion;
var a:integer;
    counter:integer;
    png:TPNGImage;
    bitmap:TBitmap;
    dc:hdc;
begin
try
  try
    hotkey := GlobalAddAtom('HotkeyF10');
    RegisterHotKey(Handle, hotkey, 0, VK_F10);
    //prepare variables
    isRecording:=true;
    counter:=0;
    normalizecliprect;

    //prepare png
    bitmap:=TBitmap.Create();
    bitmap.Width:=cliprect.Right-cliprect.Left;
    bitmap.Height:=cliprect.Bottom-cliprect.Top;
    png:=TPNGImage.Create;
    png.Assign(bitmap);
    png.CompressionLevel:=1;
    dc:=getdc(0);

    while((isRecording) and (counter<1000)) do
    begin
      //capture image
      bitblt(png.Canvas.Handle,0,0,png.Width,png.Height,dc,cliprect.Left,cliprect.Top,srccopy);
      png.SaveToFile(path+PublishedSlowMotion+'\'+rightstr(inttostr(counter+1000000),6)+'.png');
      counter:=counter+1;
      sleep(5);
      Application.ProcessMessages;
    end;

    //release variables
    releasedc(0,dc);
    png.Destroy;
    bitmap.Destroy;

    //open folder with slow motion pictures
    shellexecute(handle,'open',pchar(path+PublishedSlowMotion+'\'),nil,nil,sw_show);
  except
  end;
finally
  isRecording:=false;
  isShowingRecording:=false;
  UnRegisterHotKey(Handle, hotkey);
end;
end;

procedure TClipForm.RecordVideo;
var a:integer;
    counter:integer;
    CapturingImageStart:cardinal;
    CapturingImageStop:cardinal;
    FrameRate:integer;
    milisecondsbillance:integer;
    renderingthread:cardinal;
    cursor:TPoint;
    png:TPNGImage;
    bitmap:TBitmap;
    filename:string;
    dc:hdc;
begin
try
  try
    hotkey := GlobalAddAtom('HotkeyF10');
    RegisterHotKey(Handle, hotkey, 0, VK_F10);
    //prepare variables
    isRecording:=true;
    RecordingVideoStatus:=RecordingVideo;
    counter:=0;
    normalizecliprect;

    //12 fps
    FrameRate:=floor(1000/12);

    //prepare png
    bitmap:=TBitmap.Create();
    bitmap.Width:=cliprect.Right-cliprect.Left;
    bitmap.Height:=cliprect.Bottom-cliprect.Top;
    png:=TPNGImage.Create;
    png.Assign(bitmap);
    png.CompressionLevel:=1;
    dc:=getdc(0);

    while((isRecording) and (counter<1000000)) do
    begin
      CapturingImageStart:=gettickcount();
      //capture image
      bitblt(png.Canvas.Handle,0,0,png.Width,png.Height,dc,cliprect.Left,cliprect.Top,srccopy);
      getcursorpos(cursor);
      //test if mouse is down
      if((getasynckeystate(VK_LBUTTON)=0) and (getasynckeystate(VK_RBUTTON)=0)) then
        png.Canvas.Draw(cursor.X-cliprect.Left-(CursorImage.Width div 2),cursor.Y-cliprect.Top-(CursorImage.Height div 2),CursorImage.Picture.Graphic)
      else
        png.Canvas.Draw(cursor.X-cliprect.Left-(CursorImage.Width div 2),cursor.Y-cliprect.Top-(CursorImage.Height div 2),CursorMouseDownImage.Picture.Graphic);
      filename:=path+TempVideo+'\'+rightstr(inttostr(counter+1000000),6)+'.png';
      png.SaveToFile(filename);

      //copy images if there was long delay
      CapturingImageStop:=gettickcount();
      milisecondsbillance:=0;
      while((CapturingImageStop-(CapturingImageStart+milisecondsbillance))>=FrameRate) do
      begin
        counter:=counter+1;
        milisecondsbillance:=milisecondsbillance+83;
        copyfile(pchar(filename),pchar(path+TempVideo+'\'+rightstr(inttostr(counter+1000000),6)+'.png'),false);
        CapturingImageStop:=gettickcount();
      end;
      //sleep
      sleep(83-(CapturingImageStop-(CapturingImageStart+milisecondsbillance)));

      counter:=counter+1;
      Application.ProcessMessages;
    end;

    //release variables
    releasedc(0,dc);
    png.Destroy;
    bitmap.Destroy;
    //render video
    beginthread(nil,0,RenderVideo,nil,0,renderingthread);
  except
  end;
finally
  isRecording:=false;
  isShowingRecording:=false;
  UnRegisterHotKey(Handle, hotkey);
end;
end;

procedure TClipForm.Timer1Timer(Sender: TObject);
begin
  if(shape1.pen.color=rgb(255,0,0))then
  begin
    shape1.pen.color:=rgb(0,255,0);
    shape2.pen.color:=rgb(0,255,0);
    shape3.pen.color:=rgb(0,255,0);
    shape4.pen.color:=rgb(0,255,0);
  end
  else
  begin
    shape1.pen.color:=rgb(255,0,0);
    shape2.pen.color:=rgb(255,0,0);
    shape3.pen.color:=rgb(255,0,0);
    shape4.pen.color:=rgb(255,0,0);
  end;

  if(visible=true) then
  begin
    formstyle:=fsstayontop;
    bringtofront;
    //Application.NormalizeTopMosts;
    SetWindowPos(Handle, HWND_TOPMOST, 0, 0, 0, 0,
                 SWP_NOACTIVATE + SWP_NOMOVE + SWP_NOSIZE);
    positionshapes;
  end;
end;

procedure TClipForm.WMHotKey(var Msg: TWMHotKey);
begin
  //stop recording
  if(msg.HotKey=hotkey)then
    StopRecording;
end;

end.
